/*
   Coolpix RAW to Coolpix NEF converter
   by paul69, v0.05, 2003-01-29, <e2500@narod.ru>

   Main function

   Supported camera (tested)
      E2500
      E4500
      E995
      E950
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#pragma hdrstop
#include "tiff.h"
#include "file.h"
#include "raw2nef.h"

#define program "raw2nef v0.05, by paul69 <e2500@narod.ru> [compiled: " __DATE__ " " __TIME__ "]\n"

bool raw_to_nef(int n, char* a[]);

int main(int argc, char* argv[] )
{
   return raw_to_nef( argc, argv );
}

static void usage();
static char ofn_buf[_MAX_PATH];
static char efn_buf[_MAX_PATH];

static
TCCDParam default_pars[] = {
   {1600+16, 1200+4, 0x2CE000, "YCGM", 2424, 12, {{ 0, (1200+4)/2 }, {0x2CE000/2, (1200+4)/2}}  }, // E2500
   {2272+16, 1704+4, 0x598000, "GMYC", 3432, 12, {{ 0, (1704+4)/2 }, {0x598000/2, (1704+4)/2}}  }, // E4500
   {2048+16, 1536+4, 0x48D000, "MGCY", 3096, 12, {{ 0, (1536+4)/2 }, {0x48D000/2, (1536+4)/2}}  }, // E995
   {1600+16, 1200+4, 0x25A000, "YCGM", 2048, 10, {{ 0, (1200+4)/2 }, {0x25A000/2, (1200+4)/2}}  }, // E950
};

const int n_default_pars = sizeof(default_pars) / sizeof(default_pars[0]);

char* CreateExifFileName(const char* FilePath, int NameOffset, int ExtOffset)
{
   if( NameOffset + 8 != ExtOffset )
      return NULL;

   char* standard_name = "DSCN";

   for(int i=0; i<4; ++i)
   {
      if( (FilePath[NameOffset+i] & ~0x20) != standard_name[i] )
         return NULL;
   }

   int n = atoi( FilePath+NameOffset+4 );
   if( n <= 0 ) return NULL;

   char* efn = efn_buf;
   memcpy(efn, FilePath, NameOffset);
   sprintf(efn + NameOffset, "DSCN%04d.JPG", n+1);
   return efn_buf;
}

bool ProcessFile(const char* ifn, char* ofn, char* efn, TCCDParam const& user_pars, bool verbose, bool copy_exif, bool endian)
{
   int i;

   if( !ofn || !efn )
   {
      const char* inpFileName = ifn;

      const char* ifn_slash = strrchr(inpFileName, '/');
      if( ifn_slash )
         inpFileName = ifn_slash+1;

      const char* ifn_bslash = strrchr(inpFileName, '\\');
      if( ifn_bslash )
         inpFileName = ifn_bslash+1;

      const char* inpFileExt = inpFileExt = strrchr(inpFileName, '.');

      if( !inpFileExt )
         inpFileExt = ifn + strlen(ifn);

      if( !ofn )
      {
         ofn = ofn_buf;
         int len = inpFileExt - ifn;
         memcpy(ofn, ifn, len);
         strcpy(ofn+len, ".NEF");
      }

      if( !efn && copy_exif )
      {
         efn = CreateExifFileName(ifn, inpFileName-ifn, inpFileExt-ifn);
         if( !efn && verbose )
            printf("warning: cannot get exif filename\n");
      }
   }

   TFile in = fopen( ifn, "rb");
   if( !in ) { printf("error: \"%s\" - cannot open\n", ifn); return false; }

   fseek( in, 0, SEEK_END );
   unsigned file_size = ftell( in );

   if( verbose )
      printf("input: %s, size=%u\n", ifn, file_size);

   for(i=0; i<n_default_pars; ++i)
   {
      if( file_size == default_pars[i].file_size )
         break;
   }

   TCCDParam ccd_pars;

   if( i < n_default_pars )
   {
      if( verbose )
         printf("found default parameters for size=%u\n", file_size);

      ccd_pars = default_pars[i];
      if( user_pars.raw_width != ~0 ) ccd_pars.raw_width = user_pars.raw_width;
      if( user_pars.raw_height != ~0 ) ccd_pars.raw_height = user_pars.raw_height;
      if( user_pars.cfa_pattern[0] != -1 ) memcpy( ccd_pars.cfa_pattern, user_pars.cfa_pattern, 4);
      if( user_pars.bits_per_sample != ~0 ) ccd_pars.bits_per_sample = user_pars.bits_per_sample;
      if( user_pars.row_length != ~0 ) ccd_pars.row_length = user_pars.row_length;
      if( user_pars.field[0].offset != ~0 ) ccd_pars.field[0].offset = user_pars.field[0].offset;
      if( user_pars.field[0].rows != ~0 ) ccd_pars.field[0].rows = user_pars.field[0].rows;
      if( user_pars.field[1].offset != ~0 ) ccd_pars.field[1].offset = user_pars.field[1].offset;
      if( user_pars.field[1].rows != ~0 ) ccd_pars.field[1].rows = user_pars.field[1].rows;
   }
   else
   {
      if( verbose )
         printf("not found default parameters for size=%u\n", file_size);

      if( user_pars.raw_width == ~0 ) { printf("error: unknown raw width\n"); return false; }
      if( user_pars.raw_height == ~0 ) { printf("error: unknown raw height\n"); return false; }
      if( user_pars.cfa_pattern[0] == -1 ) { printf("error: unknown cfa_pattern\n"); return false; }
      if( user_pars.bits_per_sample == ~0 ) { printf("error: unknown bits per sample\n"); return false; }
      if( user_pars.row_length == ~0 ) { printf("error: unknown row length\n"); return false; }
      if( user_pars.field[0].offset == ~0 ) { printf("error: unknown field 0 offset\n"); return false; }
      if( user_pars.field[0].rows == ~0 ) { printf("error: unknown field 0 rows\n"); return false; }
      if( user_pars.field[1].offset == ~0 ) { printf("error: unknown field 1 offset\n"); return false; }
      if( user_pars.field[1].rows == ~0 ) { printf("error: unknown field 1 rows\n"); return false; }

      ccd_pars = user_pars;
   }

   if( verbose )
   {
      printf("current parameters:\n");
      printf("raw width: %u, raw height: %u\n", ccd_pars.raw_width, ccd_pars.raw_height);
      printf("cfa pattern: %4.4s, bits: %u\n", ccd_pars.cfa_pattern, ccd_pars.bits_per_sample );
      printf("row length: %u\n", ccd_pars.row_length);
      printf("field 0: offset: 0x%08X rows: %u\n", ccd_pars.field[0].offset, ccd_pars.field[0].rows );
      printf("field 1: offset: 0x%08X rows: %u\n", ccd_pars.field[1].offset, ccd_pars.field[1].rows );
   }

   TBuff ccd(ccd_pars.raw_height*ccd_pars.raw_width);

   if( !ccd )
   {
      printf("error: memory allocation fault (%u)\n", ccd_pars.raw_height*ccd_pars.raw_width*sizeof(color_t));
      return false;
   }

   if( verbose )
      printf("read raw data from \"%s\"\n", ifn);

   if( !read_raw( in, ccd, ccd_pars) )
   {
      printf("error: can not read raw data\n");
      return false;
   }

   in.close();

   NEF_IFD Nef_IFD;

   if( efn )
   {
      if( verbose )
         printf("read exif info from \"%s\"\n", efn);

      TFile exif = fopen( efn, "rb");

      if( !exif && verbose )
         { printf("warning: \"%s\" - cannot open exif file\n", efn); }
      else
      if( !Nef_IFD.read_exif( exif ) && verbose )
         { printf("warning: \"%s\" - cannot read exif file\n", efn); }
   }

   if( verbose )
      printf("write picture to: \"%s\"\n", ofn);

   TFile out = fopen( ofn, "wb");
   if( !out ) { printf("error: \"%s\" - can not open\n", ofn); return false; }

   if( !Nef_IFD.write_nef(out, endian, ccd, ccd_pars) )
      { printf("error: can not write NEF\n"); return false; }

   if( verbose )
      printf("OK\n");

   return true;
}

bool raw_to_nef(int n, char* a[])
{
   TCCDParam user_pars;
   memset( &user_pars, ~0, sizeof(user_pars));

   // endian: true for Intel, false for Motorolla
   bool endian = true;
   bool copy_exif = true;
   char* ifn = NULL;
   char* ofn = NULL;
   char* efn = NULL;
   bool verbose = false;
   bool batch_mode = true;

   if( n <= 1 )
   {
      usage();
      return false;
   }

   int i;
   for(i=1; i<n; ++i)
   {
      if( a[i][0] == '-' )
      {
         switch( a[i][1] )
         {
         case 'p':
            memcpy(user_pars.cfa_pattern, &a[i][3], 4);
            break;

         case 'w':
            sscanf(&a[i][3], "%i", &user_pars.raw_width);
            break;

         case 'l':
            sscanf(&a[i][3], "%i", &user_pars.row_length);
            break;

         case 'x':
            sscanf(&a[i][3], "%i", &user_pars.bits_per_sample);
            break;

         case 'h':
            sscanf(&a[i][3], "%i", &user_pars.raw_height);
            break;

         case 'b': {
            int f = atoi(&a[i][2]); if( f >= 2 ) goto invalid;
            sscanf(&a[i][4], "%i", &user_pars.field[f].offset);
            } break;

         case 'n': {
            int f = atoi(&a[i][2]); if( f >= 2 ) goto invalid;
            sscanf(&a[i][4], "%i", &user_pars.field[f].rows);
            } break;

         case 'v': verbose = true; break;

         case 'e':
            batch_mode = false;
            if( i+1 < n && a[i+1][0] != '-' )
               efn = a[++i];
            else
               goto missing;
            break;

         case 'o':
            batch_mode = false;
            if( i+1 < n && a[i+1][0] != '-' )
               ofn = a[++i];
            else
               goto missing;
            break;

         case '?': usage(); return false;

         default:
            printf("warning: unknown switch: %s\n", a[i]);
            break;

         invalid:
            printf("warning: invalid switch: %s\n", a[i]);
            break;

         missing:
            printf("warning: missing argument for %s\n", a[i]);
            break;
         }
      }
      else
      {
         if( !ifn )
            ifn = a[i];
      }
   }

   if( verbose )
      printf( program );

   if( !ifn )
   {
      printf("error: input file is missing\n\n");
      return false;
   }

   if( batch_mode )
   {
      for(i=1; i<n; ++i)
      {
         if( a[i][0] == '-' )
         {
            switch( a[i][1] )
            {
            case 'o':
            case 'e':
               if( i+1 < n && a[i+1][0] != '-' )
                  ++i;
               break;
            }
         }
         else
         {
            if( a[i][0] )
               ProcessFile(a[i], NULL, NULL, user_pars, verbose, copy_exif, endian);
         }
      }
   }
   else
   {
      ProcessFile(ifn, ofn, efn, user_pars, verbose, copy_exif, endian);
   }
   return true;
}

void usage()
{
   printf("Coolpix RAW data to E5700 NEF file converter\n");
   printf( program );

   printf("\nargs: input_file1 [input_file2...] [options]\n");

   printf("\noptions: frame pars (use 0x notation for hex)\n");
   printf("   -w:1616 - raw width\n");
   printf("   -h:1204 - raw height\n");
   printf("   -p:YCGM - CFA pattern\n");
   printf("   -l:2424 - row length\n");
   printf("   -x:12 - bits per sample\n");
   printf("   -b0:0 - begin of field 0\n");
   printf("   -n0:602 - lines in field 0\n");
   printf("   -b1:0x0167000 - begin of field 1\n");
   printf("   -n1:602 - lines in field 1\n");

   printf("\noptions: processing\n");
   printf("   -v - verbose\n");
   printf("   -e <filename> - exif jpeg file name (disable batch mode)\n");
   printf("   -o <filename> - nef output file name (disable batch mode)\n");

   printf("\nSupported cameras: E2500, E4500, E995, E950\n");
}

